import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Set;
import java.util.Map.Entry;


public class DocumentDirectory 
{
	private HashMap<Integer, Integer> directory;
	private PriorityQueue<HeapObject> docBalance;
	
	public DocumentDirectory()
	{
		directory=new HashMap<Integer, Integer>();
		docBalance=new PriorityQueue<HeapObject>();
	}
	
	public void insertDocument(int docId) 
	{
		HeapObject top=docBalance.poll();
		directory.put(docId, top.nodeid);
		docBalance.add(new HeapObject(top.nodeid,top.docCount+1));
	}
	
	public void insertDocument(int docId,int nodeId)
	{
		HeapObject[] heapToArray=new HeapObject[docBalance.size()];
		heapToArray=docBalance.toArray(heapToArray);
		directory.put(docId, nodeId);
		for(HeapObject h: heapToArray)
		{
			if(h.nodeid==nodeId)
			{
				HeapObject newObject=new HeapObject(nodeId,h.docCount+1);
				docBalance.remove(h);
				docBalance.add(newObject);
			}
		}
	}
	
	public int seachDirectory(int docId)
	{
		return directory.get(docId);
	}
	
	public ArrayList<DirectoryElement> updateDirectory(int start,int stop)
	{
		ArrayList<DirectoryElement> documents=new ArrayList<DirectoryElement>();
		Set<Map.Entry<Integer,Integer>> elements=directory.entrySet();
		Iterator <Map.Entry<Integer,Integer>> itr=elements.iterator();
		while(itr.hasNext())
		{
			Map.Entry<Integer,Integer> entry=itr.next();
			if(entry.getKey()>start && entry.getKey()<stop)
			{
				documents.add(new DirectoryElement(entry.getKey(),entry.getValue()));
			}
		}
		
		return documents;
	}
	
	public int[][] redistribute(int activeNodes)
	{	
		int docsNewNode=directory.size()/activeNodes;
		int docsPrNode=docsNewNode/(activeNodes-1);
		int[]docCount=new int[7];
		for(int i=0;i<docCount.length;i++)
		{
			docCount[i]=0;
		}
		int[][] docsToMove=new int[7][docsPrNode];
		Set<Map.Entry<Integer,Integer>> elements=directory.entrySet();
		Iterator <Map.Entry<Integer,Integer>> itr=elements.iterator();
		while(itr.hasNext())
		{
			Map.Entry<Integer,Integer> entry=itr.next();
			if(docCount[entry.getValue()-1]<docsPrNode)
			{
				docsToMove[entry.getValue()-1][docCount[entry.getValue()-1]]=entry.getKey();
				itr.remove();
			}
		}
		
		return docsToMove;
	}
	public int getDocCount()
	{
		return directory.size();
	}
	
	public void createHeap(int activeNodes)
	{
		for(int i=0;i<activeNodes;i++)
		{
			int nodenr=i+1;
			if(nodenr>3)
				nodenr++;
			ReportDocCountInterface rdc = null;
			int nodeload=0;
			try {
				rdc=(ReportDocCountInterface)Naming.lookup("rmi://endless0"+nodenr+".idi.ntnu.no/GetDocCount");
				nodeload=rdc.getDocCount(-1);
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NotBoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch(RemoteException e)
			{
				e.printStackTrace();
			}
			
			HeapObject h=new HeapObject(nodenr,nodeload);
			docBalance.add(h);
		
		}
	}
	class HeapObject implements Comparable<HeapObject>
	{
		int nodeid;
		int docCount;
		
		public HeapObject(int nodeid,int docCount)
		{
			this.nodeid=nodeid;
			this.docCount=docCount;
		}
		
		public int compareTo(HeapObject hob) 
		{
			if(hob.docCount<this.docCount)
			{
				return -1;
			}
			else if(hob.docCount>this.docCount)
			{
				return 1;
			}
			else 
			{
				return 0;
			}
			
		}
		
		
	}
	
	

}
