import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.ArrayList;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;



public class ManagerMap 
{
	private DocumentDirectory dir;
	private StorageNode sNode;
	private int ACTIVE_NODES;
	private DistributeDocuments distr;
	private int port;
	private ServerSocket server;
	
	public ManagerMap(StorageNode sNode)
	{
		dir=new DocumentDirectory();
		distr=new DistributeDocuments("rangepartition");
		this.sNode=sNode;
		ACTIVE_NODES=1;
		this.port=5555;
		try {
			server=new ServerSocket(port);
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		try {
			Naming.rebind("UpdateManager",new UpdateManagerMap(this));
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void receiveDocuments(Socket managerSocket)
	{
		System.out.println("Starting manager");
        ManagerReceiveThread receiver=new ManagerReceiveThread(managerSocket,this);
        receiver.run();
          
		
	}
	
	public void updateDirectory()
	{
		System.out.println("Update Manager Directory");
		ACTIVE_NODES=sNode.getACTIVE_NODES();
		int range=Integer.MAX_VALUE/ACTIVE_NODES;
		System.out.println("Range "+range);
		int nodeNum,rangeEnd;
		ArrayList<DirectoryElement> documentsToMove;
		ArrayList<Integer> nodedir=sNode.nodedir;
		UpdateManagerMapInterface update;
		System.out.println("Nodedir size "+nodedir.size());
		for(int i=0;i<nodedir.size();i++)
		{
			nodeNum=nodedir.get(i);
			
			if(nodeNum>3)
				nodeNum++;
			System.out.println("Nodenum "+ nodeNum);
			if(nodeNum!=sNode.getNodeNum())
			{
				rangeEnd=range*i;
				documentsToMove=dir.updateDirectory(range-range,rangeEnd);
				System.out.println("Documents to redistribute "+ documentsToMove.size());
				try {
					update=(UpdateManagerMapInterface)Naming.lookup("rmi://endless0"+nodeNum+".idi.ntnu.no/UpdateManager");
					update.addManagerElements(documentsToMove);
				} catch (MalformedURLException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
				} catch (RemoteException e) {
						// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (NotBoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				}
			}
			
		}
		addNode();
		
	}
	
	public void addNode()
	{
		System.out.println("Add Node");
		int[][] docsToMove =dir.redistribute(ACTIVE_NODES);
		XfsUpdateInterface xfsup;
		for(int j=0;j<docsToMove.length;j++)
		{
			int nodenr=j+1;
			if(nodenr==4)
				continue;
			try {
				xfsup=(XfsUpdateInterface)Naming.lookup("rmi://endless0"+nodenr+".idi.ntnu.no/XfsUpdate");
				xfsup.redistributeToNewNode(docsToMove[j]);
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (RemoteException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NotBoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
		}
		
		
	}
	
	public int getACTIVE_NODES() {
		return ACTIVE_NODES;
	}
	public void setACTIVE_NODES(int active_nodes) {
		ACTIVE_NODES = active_nodes;
	}
	
	public void addToDirectory(ArrayList<DirectoryElement> elements)
	{
		for(int i=0;i<elements.size();i++)
		{
			dir.insertDocument(elements.get(i).getKey(), elements.get(i).getNodeId());
		}
	}

	class ManagerReceiveThread implements Runnable
	{
		private Socket socket;
		private ManagerMap manager;
		String filePath="/mnt/d1/hoyum/documents";
		
		public ManagerReceiveThread(Socket socket,ManagerMap manager)
		{
			this.socket=socket;
			this.manager=manager;
			
		}

		public void run() 
		{
			BufferedReader in = null;
			File gov2File=null;
			try {
				in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			String docTitle,docUrl,docContent;
			int contentLength;
			try {
				boolean listening=true,reading=false;
				
				while(listening)
				{
				
					if((in.read()!=-1))
					{
						System.out.println("Manager reading file ");
						reading=true;

	
						docTitle=in.readLine();

						contentLength=Integer.parseInt(in.readLine());

						docUrl=in.readLine();
					
					

						int c;
						ArrayList<Character> inputArray=new ArrayList<Character>();
						for(int i=0;i<contentLength;i++)
						{
							c=in.read();
							Character element=new Character((char)c);
							inputArray.add(element);
						}
						inputArray.trimToSize();
						char[] charArray=new char[inputArray.size()];
				
						for(int i=0;i<inputArray.size();i++)
						{
						charArray[i]=inputArray.get(i).charValue();
						}
						inputArray=null;
						docContent=new String(charArray);
						int docId=Math.abs(docUrl.hashCode());
						Document doc=createDocument(docContent,docUrl,docTitle);
						distr.initRangepartitioning(sNode.getNodedir(), sNode.getACTIVE_NODES());
						int node=distr.findNode(docId);
						System.out.println("Range check "+ node);
						if(node!=sNode.getNodeNum())
						{
							sNode.sendDocument(doc, docContent, node,true);
						}
						else
						{
							dir.createHeap(sNode.getACTIVE_NODES());
							dir.insertDocument(docId);
							System.out.println("Inserting to manager map");
							node=checkDirectory(docId);
							if(node==sNode.getNodeNum())
							{
								sNode.loggingDocCount++;
								gov2File=new File(filePath+"/"+docTitle+".txt");
								
								FileWriter writer=new FileWriter(gov2File);
								writer.write(docContent);
							
							
								writer.close();
							

							
								indexDocument(doc);
							 
							
								System.out.println("Retrieved docs"+ sNode.docCount());
							}
							else
							{	
								sNode.sendDocument(doc, docContent, node,false);
							}
							System.out.println("Nodeid from directory "+ node);
							reading=false;
						}
						

						
					}
					
				}//end while
			
			}//end try-block
			catch(IOException ioe)
			{
				System.out.println(ioe.getMessage());
				
			}
		}
		
		private int checkDirectory(int docId)
		{
			return manager.dir.seachDirectory(docId);
			
		}
		
		private Document createDocument(String docContent, String docUrl,
				String docTitle) {
			Document doc=new Document();
			   
			   doc.add(new Field("url",docUrl,Field.Store.COMPRESS,Field.Index.NOT_ANALYZED));
			   doc.add(new Field("docId",new Integer(Math.abs(docUrl.hashCode())).toString(),Field.Store.COMPRESS,Field.Index.NOT_ANALYZED));
			   doc.add(new Field("title",docTitle,Field.Store.COMPRESS,Field.Index.NOT_ANALYZED));
			   doc.add(new Field("content",docContent,Field.Store.NO,Field.Index.ANALYZED));
			return doc;
		}
			
	
		
		
	}

	   public synchronized void indexDocument(Document doc)
	   {
		   try 
		   {
			   IndexWriter iWriter=sNode.getIndexWriter();

			   iWriter.addDocument(doc);
			   
			   sNode.increaseDocCount();
		   }
		   catch (CorruptIndexException e) 
		   {
			   System.out.println(e.getMessage());
			   e.printStackTrace();
		   }
		   catch (IOException e) 
		   {
			   System.out.println(e.getMessage());
			   e.printStackTrace();
		   }
	   }

	public ServerSocket getServer() {
		return server;
	}
}
