
/*******************
 * Main class running at the nodes. Monitors node state and 
 * performs the redistribution
 *********************/

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.rmi.NoSuchObjectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;



import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.LockObtainFailedException;



public class StorageNode 
{
	private int DOCLIMIT; //the max number of documents node can store
	private int ACTIVE_NODES;
	private String method;

	private int docCount; // the number of documents currently stored at the node
	private String indexPath="/mnt/d1/hoyum/index";
	private IndexWriter iWriter;
	private IndexReader iReader;
	private String NODE_NAME;
	private boolean repartitioning=false;
	private ServerSocket serverSocket;
	private SendDocument sender;
	private SendDocument managerSender;
	
	//Variables specific for different methods;
	private int clusterLevel;
	private long timer;
	private CacheTable cTable;
	private FileWriter logg; 
	public int loggingDocCount;
	public ArrayList<Integer> nodedir;
	private ManagerMap manager;
	private boolean withIndex;
	
	
	


	public void setClusterLevel(int clusterLevel) {
		this.clusterLevel = clusterLevel;
	}

	/******
	 * Constructors
	 **********/
	
	public StorageNode(String method,boolean withIndex) throws RemoteException
	{
		this.method=method;
		this.withIndex=withIndex;
		DOCLIMIT=125000;
		ACTIVE_NODES=1;
		docCount=0;
		clusterLevel=0;
		cTable=new CacheTable();
		loggingDocCount=0;
		timer=-1;
		try {
			Calendar c=Calendar.getInstance();
			String currentTime=c.get(Calendar.DATE)+"";
			logg = new FileWriter(new File("/mnt/d1/hoyum/logg"+currentTime));
			logg.write("Indexing time"+"\n");
			iWriter=new IndexWriter(indexPath,new StandardAnalyzer(),true,IndexWriter.MaxFieldLength.LIMITED);
			InetAddress addr = InetAddress.getLocalHost();
			NODE_NAME=addr.getHostName();
			iReader=IndexReader.open(indexPath);
			sender=null;
			managerSender=null;
			//Binding remote methods
			Naming.rebind("Scaling",new AddNode(this));
			Naming.rebind("GetDocCount",new ReportDocCount(this));
			Naming.rebind("UpdateClusterImage", new UpdateClusterImage(this));
			Naming.rebind("XfsUpdate",new XfsUpdate(this));
			
			//Variables speciffic for different methods
			nodedir=new ArrayList<Integer>();
			nodedir.add(0,1);
		} catch (CorruptIndexException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (LockObtainFailedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	


	/*******
	 * Returns the hostname of the current node
	 * @return
	 */
	public int getNodeNum()
	{
		return Integer.parseInt(new Character(NODE_NAME.charAt(8)).toString());
		
	}
	
	public int getACTIVE_NODES() {
		return ACTIVE_NODES;
	}
	
	public void setACTIVE_NODES(int active_nodes) {
		ACTIVE_NODES = active_nodes;
	}


	public String getMethod() {
		return method;
	}

	public int getClusterLevel() {
		return clusterLevel;
	}

	public CacheTable getCTable() {
		return cTable;
	}

	public void setCTable(CacheTable table) {
		cTable = table;
		System.out.println(" New maxLevel "+cTable.maxLevel);
	}

	public ArrayList<Integer> getNodedir() {
		return nodedir;
	}
	/**********
	 * Uses the disrtribution algorithms to 
	 * calculate the hash key for the documents that
	 * is to be transported to new nodes
	 * @return a matrix of doc keys pr node
	 **********/
	@SuppressWarnings("unchecked")
	private ArrayList<Document>[] docsForTransport()
	{
		ArrayList<Document> [] docKeys=new ArrayList[ACTIVE_NODES];
		System.out.println("Active Nodes "+ ACTIVE_NODES);
		//docKeys=new Document[ACTIVE_NODES][];
		int[] ArraydocCount=new int[6];
		
		try 
		{
			IndexWriter iWriter=getIndexWriter();
			iWriter.optimize();
			iWriter.commit();
			//int numDocs=iWriter.numDocs();
			

			
			//instansiate values of various arrays
			for(int nodes=0;nodes<6;nodes++)
			{
				//docKeys=new Document[nodes][numDocs];
				if(nodes<ACTIVE_NODES)
				{
					docKeys[nodes]=new ArrayList<Document>();
				}
				ArraydocCount[nodes]=0;
			}
			
			docKeys=partitionDocuments(docKeys, ArraydocCount);
			System.out.println("Returning the document array");
			
		}
		catch (CorruptIndexException e) 
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		catch (IOException e) 
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		
		return docKeys;
	}

	/****************
	 * Distribute the documents after standard range-partitioning
	 * @param docKeys
	 * @param docCount
	 * @param iReader
	 * @param numDocs
	 * @throws CorruptIndexException
	 * @throws IOException
	 *******************/
	private ArrayList<Document>[] partitionDocuments(ArrayList<Document>[] docKeys, int[] ArraydocCount) throws 
			IOException {
		//Distribute docs after standard range partitioning
		IndexReader reader=IndexReader.open(FSDirectory.getDirectory(indexPath),true);
		int numDocs=reader.numDocs();
		DistributeDocuments distr=new DistributeDocuments(method);
		
		if(method.equalsIgnoreCase("lh"))
		{
			distr.initLH(ACTIVE_NODES, clusterLevel);
		}
		else if(method.equalsIgnoreCase("eh"))
		{
			distr.initEH(cTable);
		}
		else if(method.equalsIgnoreCase("rush"))
		{
			distr.initRush(ACTIVE_NODES);
		}
		else if(method.equalsIgnoreCase("rangepartition"))
		{
			distr.initRangepartitioning(nodedir, ACTIVE_NODES);
		}
		else if(method.equalsIgnoreCase("scaddar"));
		{
			distr.initSCADDAR(ACTIVE_NODES);
		}
		

		System.out.println("iReader num docs "+numDocs);
		logg.write("Num docs on node " + numDocs);
		for(int docItr=0;docItr<reader.maxDoc();docItr++)
		{
			if(reader.isDeleted(docItr))
				continue;
			int nodeId=Integer.parseInt(reader.document(docItr).get("docId"));
			//distr.findNode(nodeId);
			int node=distr.findNode(nodeId);
			/**if(node>ACTIVE_NODES)
			{
				node=node%ACTIVE_NODES;
				node++;
			}**/
			if(node>3)
				node++;
			if(node!=getNodeNum())
			{
				System.out.println("Active nodes " + ACTIVE_NODES);
				System.out.println("Returned node id "+ node);
				sendDocument(reader.document(docItr),null,node,false);
			}
			else
			{
				continue;
			}
			//System.out.println("DocKeys index "+ node);
			//docKeys[node-1].add(reader.document(docItr));

			if((node-1)<ArraydocCount.length)
			{
				ArraydocCount[node-1]++;
				
			}
			/*if(node>3)
				node++;
			
			if(node!=getNodeNum())
			{
				sendDocument(reader.document(docItr),null,node);
			}*
			/*boolean partitioning=true;
			while(partitioning)
			{
				
				if(nodeId<((Integer.MAX_VALUE/ACTIVE_NODES)*(node)))
				{
					docKeys[node-1].add(reader.document(docItr));
					ArraydocCount[node-1]++;
					partitioning=false;
				}
				node++;
			}
			*/
			
			
		}
		System.out.println("Node dir");
		for(int i=0;i<nodedir.size();i++)
		{
			System.out.print(" "+nodedir.get(i)+" ");
		}
		System.out.println("");
		int localNode=getNodeNum();
		System.out.println("local node "+localNode);
		//Setting new docCount
		int newDocCount=0;
		for (int k=0;k<ArraydocCount.length;k++)
		{
			if( k==(localNode-1))
			{
				continue;
			}
			
			System.out.println("Sending to node "+k + " " + ArraydocCount[k] );
			newDocCount+=ArraydocCount[k];
			
		}
		setDocCount(docCount-newDocCount);
		int remainingDocs=docCount-newDocCount;
		iReader=IndexReader.open(FSDirectory.getDirectory(indexPath),true);
		reader.reopen();
		logg.write(" New docCount " +docCount()+"\n");
		System.out.println("New doc count " + iReader.numDocs());
		reader.close();
		return docKeys;
	}
	
	
	/*******
	 * Returns an IndexWriter that all classes that writers to the index shall use. 
	 * @return a joint IndexWriter
	 ********/
	public IndexWriter getIndexWriter()
	{
		return iWriter;
	}
	
	/***********
	 * Returns an IndexReader for the index located at the node
	 * @return
	 *************/
	public IndexReader getIndexReader()
	{
		return iReader;
	}
	public void setIReader() {
		try {
			iReader=IndexReader.open(FSDirectory.getDirectory(indexPath),true);
		} catch (CorruptIndexException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/************
	 * Closes the joint IndexWriter
	 */
	public void closeWriter()
	{
		System.out.println("Closing writer");
		try {
			iWriter.close();
		} catch (CorruptIndexException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	/**********************
	 * Increases the counter for the number of documents on the node
	 *******************/
	public void increaseDocCount()
	{
		docCount++;
		if(docCount()>=DOCLIMIT) // for hash partitioning add check for isPartitioning
		{
			setPartitioning(true);
			System.out.println("Doclimit reached ");
			//setPartitioning(true);
			try
			{
				AddNodeMonitorServerInterface ad = (AddNodeMonitorServerInterface)Naming.lookup("rmi://129.241.102.111/Scaling");
				ad.addNode(this.getNodeNum());
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (RemoteException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (NotBoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	/**************
	 * Decreases the document counter
	 **************/
	public void decreaseDocCount()
	{
		docCount--;
	}
	
	/******
	 * Retreives the number value of the docCount variable
	 * @param args
	 ******/
	
	public int docCount()
	{
		IndexWriter writer=getIndexWriter();
		int numDocs = 0;
		try {
			writer.commit();
			numDocs=writer.numDocs();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return numDocs;
		
	}
	
	/*******
	 * Sets a new value for docCount
	 * @param args
	 **********/
	
	public void setDocCount(int count)
	{
		docCount=count;
	}
	
	public long getTimer() {
		return timer;
	}

	public void setTimer(long timer) {
		this.timer = timer;
	}

	/**************
	 * Sets the state of the node to repartitioning
	 * @param args
	 *******/
	
	public void setPartitioning(boolean state)
	{
		repartitioning=state;
	}
	
	/******************
	 * Returns the current state of the node regarding repartitioning
	 * @return
	 */
	public boolean isPartitioning()
	{
		return repartitioning;
	}
	
	

	public static void main(String[] args)
	{
		StorageNode nodeInstance = null;
		try {
			nodeInstance = new StorageNode("rangepartition",true);
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println("Node number "+ nodeInstance.getNodeNum());
		nodeInstance.receiveDocuments();
		//nodeInstance.closeResources();

	}// end main

	/**
	 * Method called when the node is full
	 * Begins the process of activating the scaling process
	 */
	private void distributeDocuments() {

		System.out.println("DocCount "+docCount);

		try {
			iReader=IndexReader.open(indexPath);
		} catch (CorruptIndexException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if(docCount==0)
			return;
		ArrayList<Document>[] docsForTransport=docsForTransport();
		int localNode=getNodeNum();
		System.out.println("docsForTransport "+ docsForTransport.length);

		for(int i=0;i<docsForTransport.length;i++)
		{
			int nodeId=i+1;
			int newNode=(int)Math.pow(2,clusterLevel-1);
			if(nodeId!=(localNode+newNode))
				continue;
			if(nodeId>3)
				nodeId++;
			if(nodeId!=localNode)
			{
				System.out.println("NodeId "+ nodeId);
				
				//docsForTransport[i].trimToSize();
				System.out.println("Size arraylist "+ docsForTransport[i].size());
				docsForTransport[i].trimToSize();
				for(int k=0; k<docsForTransport[i].size();k++)
				{	
					//sendDocument(docsForTransport[i].get(k),null,nodeId);

				}
				System.out.println("Sending done");
				
			}
			
		}// end for-loop
	}
	
	
	/**********
	 * Method that handles the IO through the socket interface
	 * @param withIndex
	 */
    public void receiveDocuments()
    {
        
    	manager=new ManagerMap(this);
    	
        int port=4444;
        boolean listening = true;


        try {
            serverSocket = new ServerSocket(port);

            while (listening)
            {            	
            	//manager.receiveDocuments(manager.getServer().accept());
            	ReceiveDocumentsThread receiver=new ReceiveDocumentsThread(serverSocket,this,withIndex);
            	
            	receiver.run();
            	
            }
            

        } catch (IOException e) {
            System.err.println("Could not listen on port: "+port);
        }
        finally
        {
        	try {
				logg.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        }
       
    }// receiveDocuments

	/*****************
	 * Method Starts the repartitioning of the cluster
	 ***************/
	private void repartitionCluster() {
		try {
			/*for(int i=1;i<=ACTIVE_NODES;i++)
			{
				if(i>3)
					i+=1;
				
				AddNodeInterface ad = (AddNodeInterface)Naming.lookup("rmi://endless0"+i+".idi.ntnu.no/Scaling");
				ad.addNode();
			}*/
			AddNodeInterface ad = (AddNodeInterface)Naming.lookup("rmi://129.241.102.111/Scaling");
			ad.addNode();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotBoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void addNode()
	{	
		
		setPartitioning(true);
		
		if(docCount==0)
		{
			setPartitioning(false);
			return;
		}

		System.out.println("Scaling cluster");
		setTimer(new Date().getTime());
		distributeDocuments();
		
		calculateTime();
		
		
		setPartitioning(false);
		
		

		// TODO Auto-generated method stub
		
	}
	
	public void closeResources()
	{
		//java.rmi.server.UnicastRemoteObject.unexportObject( _rmiregistry, true );
		closeWriter();
		try {
			serverSocket.close();
			sender.closeConnection();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
	public void sendDocument(Document doc, String content, int node,boolean toManager)
	{

		if(toManager)
		{
			if(managerSender==null)
			{
				managerSender=new SendDocument(5555,this);
			}
			
			synchronized(managerSender)
			{
				managerSender.send(doc, content, node);
			}
		}
		else
		{
			if(sender==null)
			{
				sender=new SendDocument(4444,this);
			}

			
			synchronized(sender)
			{
				if(withIndex==true)
				{
					sender.sendWithIndex(doc, content, node,null,false);
				}
				else
				{
				sender.send(doc,content,node);
				}
			}
		}
	}
	
	public void calculateTime()
	{
		long stopTime=new Date().getTime();
		long time=stopTime-getTimer();
		System.out.println("Timer docDocCount " + docCount);
		System.out.println("Time " + time);
		
		try {
			String loggData=""+ACTIVE_NODES+" :"+time + " Num Docs :"+loggingDocCount;
			loggingDocCount=0;
			System.out.println(loggData);
			logg.write(loggData+"\n");
			logg.flush();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		setTimer(new Date().getTime());
		
	}

	public void splitNode(int oldNode,int newNode) 
	{
		if(oldNode>3)
			oldNode--;
		cTable.splitNode(oldNode, newNode);
		try {
			System.out.println("Update buddy node"+ newNode);
			UpdateClusterImageInterface uci;
			int node=newNode;
			if(newNode>3)
				node++;
			
			uci = (UpdateClusterImageInterface)Naming.lookup("rmi://endless0"+node+".idi.ntnu.no/UpdateClusterImage");
			uci.updateBuddyNode(cTable);
			uci = (UpdateClusterImageInterface)Naming.lookup("rmi://129.241.102.111/UpdateClientImage");
			uci.updateCacheTable(oldNode, newNode);
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotBoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void updateNodeTable(ArrayList<Integer> dir)
	{
		System.out.println("Updating node directory");
		nodedir=dir;
		/**nodedir.trimToSize();
		
		if(nodedir.size()==1)
		{
			nodedir.add(nodeid);
		}
		else
		{
			nodedir.add(nodedir.size()/2,nodeid);
		}
		//nodedir.add(nodeid);**/
		
	}
	
	public void updateManager()
	{
		manager.updateDirectory();
	}
	
	public void redistributedToNewNode(int[] docArray)
	{
		//managerSender.closeConnection();
		for(int docid:docArray)
		{
			System.out.println("Doc key "+docid);
			IndexReader reader = null;
			try {
				reader = IndexReader.open(FSDirectory.getDirectory(indexPath),true);
			} catch (CorruptIndexException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			Document doc=getDocument(new Integer(docid).toString(),reader);
			int receivingNode=ACTIVE_NODES;
			if(receivingNode>3)
				receivingNode++;
			sendDocument(doc,null,receivingNode,false);
		}
	}
	
	public Document getDocument(String docId,IndexReader iReader)
	{
		QueryParser docIdQuery=new QueryParser("docId",new StandardAnalyzer());
		Searcher searcher=new IndexSearcher(iReader);
		System.out.println("Size index "+iReader.numDocs());
		Query query=null;
		TopDocs hits=null;
		int docnr;
		query=parseQuery(docIdQuery,docId);
		System.out.println("Doc Id "+docId);
		Document doc = null;
		try {
			hits=searcher.search(query,1);
			docnr=hits.scoreDocs[0].doc;
			doc = iReader.document(docnr);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return doc;
		
	}
	
	private Query parseQuery(QueryParser queryParser, String querystr) {
		Query query = null;
		try {
			query = queryParser.parse(querystr);
		} catch (ParseException e) {
			System.out.println("Error parsing query " 
					+ querystr + ": " + e.getMessage());
			return null;
		}
		return query;
	}


}
